theory Logic 
imports Nominal Name HOAS
begin


constdefs cPROP :: string
"cPROP == ''o''"

constdefs prop_ty :: "hoas_ty"
"prop_ty \<equiv> BASE cPROP"

lemmas prop_ty_simp[simp]= prop_ty_def

lemma [eqvt]: 
  fixes p::"var prm"
  shows "(p \<bullet> prop_ty) = prop_ty"
  by (simp add: prop_ty_def perm_string eqvts)

lemma prop_ty_fresh:
  fixes x::"var"
  shows "x \<sharp> prop_ty"
  apply(simp add:prop_ty_def)
  done


nominal_datatype hoas_form =
    ATOM "hoas"
  | IMP "hoas_form" "hoas_form"
  | FORALL "hoas_ty" "\<guillemotleft>var\<guillemotright>hoas_form"

abbreviation
  FORALL_syn::"var \<Rightarrow> hoas_ty \<Rightarrow> hoas_form \<Rightarrow> hoas_form" ("\<Pi>[_:_]._" [100,100,100] 100)
where 
  "\<Pi>[x:\<tau>].\<phi> \<equiv> FORALL \<tau> x \<phi>"

abbreviation
  IMP_syn::"hoas_form \<Rightarrow> hoas_form \<Rightarrow> hoas_form" ("_ \<rightarrow> _" [80,80] 80)
where 
  "\<phi> \<rightarrow> \<psi> \<equiv> IMP \<phi> \<psi>"


constdefs pi_fun :: "hoas_ty \<Rightarrow> (hoas \<Rightarrow> hoas_form) \<Rightarrow> hoas_form" ("FORALL' _ _" [100,100] 100)
"FORALL' \<tau> f == fresh_fun (\<lambda> x. \<Pi>[x:\<tau>].(f (VAR x)))"


nominal_primrec 
subst_form :: "hoas_form \<Rightarrow> var \<Rightarrow> hoas \<Rightarrow> hoas_form" ("_{_::=_}" [100,100,100] 100)
where
  "(ATOM A){y::=t} = ATOM (A{y::=t})"
| "(IMP \<phi>1 \<phi>2){y::=t} = IMP (\<phi>1{y::=t}) (\<phi>2{y::=t})"
| "x\<sharp>(y,t) \<Longrightarrow> (FORALL \<tau> x t0){y::=t} = FORALL \<tau> x (t0{y::=t})"
  apply(finite_guess+)
  apply(rule TrueI)+
  apply(simp_all add:abs_fresh fresh_prod fresh_ty)
  apply(fresh_guess)+
  done

lemma subst_form_eqvt[eqvt]:
  fixes pi:: "var prm"
  and t1:: "hoas_form"
  shows "pi\<bullet>(t1{b::=t2}) = (pi\<bullet>t1){(pi\<bullet>b)::=(pi\<bullet>t2)}"
  apply(nominal_induct t1 avoiding: b t2 rule: hoas_form.strong_induct)
  apply(auto simp add: perm_bij fresh_prod fresh_atm fresh_bij subst_eqvt)
  done






types hoas_thy = "hoas_form list"


inductive wf_form :: "hoas_sig \<Rightarrow> hoas_ctx \<Rightarrow> hoas_form \<Rightarrow> bool"
  ("_,_ \<turnstile> _ : o" [60,60,60] 60)
where
  wff_ATOM : "\<Sigma>, \<Gamma> \<turnstile> t : prop_ty \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> ATOM t : o"
| wff_IMP : "\<lbrakk>\<Sigma>, \<Gamma> \<turnstile> \<phi> : o; \<Sigma>, \<Gamma> \<turnstile> \<psi> : o\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> \<phi> \<rightarrow> \<psi>  : o"
| wff_FORALL : "\<lbrakk>\<Sigma>, (x,\<tau>)#\<Gamma> \<turnstile> \<phi> : o; x\<sharp> \<Gamma>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> \<Pi>[x:\<tau>].\<phi> : o"

equivariance wf_form

nominal_inductive wf_form
  apply(simp_all add:fresh_sig abs_fresh)
  done

inductive_cases wf_form_inv:
"\<Sigma>,\<Gamma> \<turnstile> ATOM t : o"
"\<Sigma>,\<Gamma> \<turnstile> \<phi> \<rightarrow> \<psi>  : o"
"\<Sigma>,\<Gamma> \<turnstile> \<Pi>[x:\<tau>].\<phi> : o"


lemma wf_form_weakening:
  "\<lbrakk>\<Sigma>,\<Gamma> \<turnstile> \<Phi> : o; \<Gamma>' extends \<Gamma>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>' \<turnstile> \<Phi> : o"
  apply(nominal_induct \<Sigma> \<Gamma> \<Phi> avoiding: \<Gamma>' rule: wf_form.strong_inducts)
(* case 1 *)
  apply(blast intro: wf_form.intros weakening)
(* case 2 *)
  apply(blast intro: wf_form.intros)
(* case 3 *)
  apply(rule wf_form.intros)
  apply(subgoal_tac "(x, \<tau>) # \<Gamma>' extends (x, \<tau>) # \<Gamma>")
  apply(auto)
  done
  
lemma wf_form_subst:
  "\<lbrakk>\<Sigma>,\<Gamma>'@[(x,\<tau>)]@\<Gamma> \<turnstile> \<phi> : o; \<Sigma>,\<Gamma> \<turnstile> t : \<tau>; valid_ctx (\<Gamma>'@[(x,\<tau>)]@\<Gamma>) \<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>'@\<Gamma> \<turnstile> \<phi>{x::=t} : o"
  apply(nominal_induct \<Sigma> \<Gamma>\<equiv>"\<Gamma>'@[(x,\<tau>)]@\<Gamma>" \<phi> avoiding: \<Gamma>' \<Gamma> x t \<tau> rule: wf_form.strong_inducts)
  (* case 1 *)
  apply(clarsimp)
  apply(rule wf_form.intros)
  apply(rule substitution_general)
  apply(simp,assumption)
  apply(simp)
(* case 2 *)
  apply(clarsimp)
  apply(blast intro:wf_form.intros)
  (* case 3 *)
  apply(clarsimp)
  apply(rule wf_form.intros)
  apply(drule_tac x="(x, \<tau>) # \<Gamma>'" in meta_spec)
  apply(auto)
  apply(simp add:fresh_list_append)
  done

lemma wf_form_subst1:
  "\<lbrakk>\<Sigma>,(x,\<tau>)#\<Gamma> \<turnstile> \<phi> : o; \<Sigma>,\<Gamma> \<turnstile> t : \<tau>; valid_ctx (\<Gamma>); x \<sharp> \<Gamma>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> \<phi>{x::=t}  : o"
  apply(subgoal_tac "\<Sigma>,[]@\<Gamma> \<turnstile> \<phi>{x::=t} : o",simp)
  apply(rule wf_form_subst,auto)
  done



inductive wf_thy :: "hoas_sig \<Rightarrow> hoas_ctx \<Rightarrow> hoas_thy \<Rightarrow> bool"
  ("_,_ \<turnstile> _ : thy" [60,60,60] 60)
where
  wft_nil : "\<Sigma>,\<Gamma> \<turnstile> [] : thy"
| wft_cons : "\<lbrakk>\<Sigma>, \<Gamma> \<turnstile> \<phi> : o; \<Sigma>, \<Gamma> \<turnstile> \<Phi> : thy\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> \<phi>#\<Phi>  : thy"


lemma wf_thy_weakening:
  "\<lbrakk>\<Sigma>,\<Gamma> \<turnstile> \<Phi> : thy; \<Gamma>' extends \<Gamma>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>' \<turnstile> \<Phi> : thy"
  apply(induct rule:wf_thy.inducts)
  apply(auto intro!:wf_thy.intros, erule wf_form_weakening,auto)
  done



inductive wf_equiv_form :: "hoas_sig \<Rightarrow> hoas_ctx \<Rightarrow> hoas_form \<Rightarrow> hoas_form \<Rightarrow> bool"
  ("_,_ \<turnstile> _ \<approx> _ : o" [60,60,60,60] 60)
where
  wf_equiv_ATOM : "\<Sigma>, \<Gamma> \<turnstile> t \<approx> u : prop_ty \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> ATOM t \<approx> ATOM u : o"
| wf_equiv_IMP : "\<lbrakk>\<Sigma>, \<Gamma> \<turnstile> \<phi> \<approx> \<phi>' : o; \<Sigma>, \<Gamma> \<turnstile> \<psi> \<approx> \<psi>' : o\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> \<phi> \<rightarrow> \<psi> \<approx> \<phi>' \<rightarrow> \<psi>' : o"
| wf_equiv_FORALL : "\<lbrakk>\<Sigma>, (x,\<tau>)#\<Gamma> \<turnstile> \<phi> \<approx> \<phi>' : o; x\<sharp> \<Gamma>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> \<Pi>[x:\<tau>].\<phi> \<approx> \<Pi>[x:\<tau>].\<phi>': o"

equivariance wf_equiv_form

nominal_inductive wf_equiv_form
  apply(simp_all add:fresh_sig abs_fresh)
  done



lemma wf_equiv_form_refl:
  "\<Sigma>,\<Gamma> \<turnstile> \<phi> : o \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> \<phi> \<approx> \<phi> : o"
  apply(nominal_induct \<Sigma> \<Gamma> \<phi> rule: wf_form.strong_inducts)
  by(auto intro!:wf_equiv_form.intros  wf_equiv_refl )




lemma wf_equiv_form_weakening:
  "\<lbrakk>\<Sigma>,\<Gamma> \<turnstile> \<Phi> \<approx> \<Phi>' : o; \<Gamma>' extends \<Gamma>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>' \<turnstile> \<Phi> \<approx> \<Phi>' : o"
  apply(nominal_induct avoiding: \<Gamma>' rule: wf_equiv_form.strong_induct)
(* case 1 *)
  apply(rule wf_equiv_form.intros)
  apply(erule wf_equiv_weakening,simp)
  (*case 2 *)
  apply(blast intro:wf_equiv_form.intros)
  (* case 3 *)
  apply(rule wf_equiv_form.intros)
  apply(subgoal_tac "(x, \<tau>) # \<Gamma>' extends (x, \<tau>) # \<Gamma>")
  apply(auto)
  done



lemma wf_equiv_form_subst:
  "\<lbrakk>\<Sigma>,\<Gamma>'@[(x,\<tau>)]@\<Gamma> \<turnstile> \<phi> \<approx> \<phi>' : o; \<Sigma>,\<Gamma> \<turnstile> t : \<tau>; valid_ctx (\<Gamma>'@[(x,\<tau>)]@\<Gamma>) \<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>'@\<Gamma> \<turnstile> \<phi>{x::=t} \<approx> \<phi>'{x::=t} : o"
  apply(nominal_induct \<Sigma> \<Gamma>\<equiv>"\<Gamma>'@[(x,\<tau>)]@\<Gamma>" \<phi> \<phi>' avoiding: \<Gamma>' \<Gamma> x t \<tau> rule: wf_equiv_form.strong_inducts)
  (* case 1 *)
  apply(clarsimp)
  apply(rule wf_equiv_form.intros)
  apply(rule wf_equiv_subst)
  apply(simp,assumption)
  apply(simp)
(* case 2 *)
  apply(clarsimp)
  apply(blast intro:wf_equiv_form.intros)
  (* case 3 *)
  apply(clarsimp)
  apply(rule wf_equiv_form.intros)
  apply(drule_tac x="(x, \<tau>) # \<Gamma>'" in meta_spec)
  apply(auto)
  apply(simp add:fresh_list_append)
  done

lemma wf_equiv_form_subst1:
  "\<lbrakk>\<Sigma>,(x,\<tau>)#\<Gamma> \<turnstile> \<phi> \<approx> \<phi>' : o; \<Sigma>,\<Gamma> \<turnstile> t : \<tau>; valid_ctx (\<Gamma>); x \<sharp> \<Gamma>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> \<phi>{x::=t} \<approx> \<phi>'{x::=t} : o"
  apply(subgoal_tac "\<Sigma>,[]@\<Gamma> \<turnstile> \<phi>{x::=t} \<approx> \<phi>'{x::=t} : o",simp)
  apply(rule wf_equiv_form_subst,auto)
  done


inductive wf_equiv_thy :: "hoas_sig \<Rightarrow> hoas_ctx \<Rightarrow> hoas_thy \<Rightarrow> hoas_thy \<Rightarrow> bool"
  ("_,_ \<turnstile> _ \<approx> _ : thy" [60,60,60,60] 60)
where
  wfet_nil : "\<Sigma>,\<Gamma> \<turnstile> [] \<approx> [] : thy"
| wfet_cons : "\<lbrakk>\<Sigma>, \<Gamma> \<turnstile> \<phi> \<approx> \<phi>' : o; \<Sigma>, \<Gamma> \<turnstile> \<Phi> \<approx> \<Phi>' : thy\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> \<phi>#\<Phi> \<approx> \<phi>'#\<Phi>'  : thy"



lemma wf_equiv_thy_refl:
  "\<Sigma>,\<Gamma> \<turnstile> \<Phi> : thy \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> \<Phi> \<approx> \<Phi> : thy"
  apply(induct rule:wf_thy.induct)
  apply(auto intro:wf_equiv_thy.intros wf_equiv_form_refl)
  done
  
lemma wf_equiv_thy_weakening:
  "\<lbrakk>\<Sigma>,\<Gamma> \<turnstile> \<Phi> \<approx> \<Phi>' : thy; \<Gamma>' extends \<Gamma>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>' \<turnstile> \<Phi> \<approx> \<Phi>' : thy"
  apply(induct rule:wf_equiv_thy.induct)
  apply(auto intro!:wf_equiv_thy.intros )
  apply(erule wf_equiv_form_weakening,auto)
  done




inductive solve :: "hoas_sig \<Rightarrow> hoas_ctx \<Rightarrow> hoas_thy \<Rightarrow> hoas_form \<Rightarrow> bool"
  ("_,_,_ \<rightharpoonup> _" [60,60,60,60] 60)
and atomic_solve :: "hoas_sig \<Rightarrow> hoas_ctx \<Rightarrow> hoas_thy \<Rightarrow> hoas_form \<Rightarrow> hoas \<Rightarrow> bool"
  ("_,_,_ | _ \<rightharpoonup> _" [60,60,60,60,60] 60)
where
  solve_ATOM: "\<lbrakk>\<phi> \<in> set \<Phi>; \<Sigma>,\<Gamma>,\<Phi> | \<phi> \<rightharpoonup> A\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup> ATOM A"
| solve_IMP: "\<Sigma>,\<Gamma>,\<phi>#\<Phi> \<rightharpoonup> \<psi> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup> \<phi> \<rightarrow> \<psi>"
| solve_FORALL: "\<lbrakk>\<Sigma>,(x,\<tau>)#\<Gamma>,\<Phi> \<rightharpoonup> \<phi>;x \<sharp> (\<Phi>,\<Gamma>)\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup> \<Pi>[x:\<tau>].\<phi>"
| asolve_ATOM: "\<lbrakk> \<Sigma>,\<Gamma> \<turnstile> A \<approx> B : prop_ty\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> | ATOM A \<rightharpoonup> B"
| asolve_IMP: "\<lbrakk>\<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup> \<phi>; \<Sigma>,\<Gamma>,\<Phi> | \<psi> \<rightharpoonup> A\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> | \<phi> \<rightarrow> \<psi> \<rightharpoonup> A"
| asolve_FORALL: "\<lbrakk>\<Sigma>,\<Gamma> \<turnstile> t : \<tau>; \<Sigma>,\<Gamma>,\<Phi> | \<phi>{x::=t} \<rightharpoonup> A; x \<sharp> (\<Gamma>,\<Phi>,A)\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> | \<Pi>[x:\<tau>].\<phi> \<rightharpoonup> A"

equivariance solve

nominal_inductive solve
  apply(simp_all add:fresh_sig abs_fresh)
  done


lemma in_closed_under_equiv:
  "\<lbrakk>\<Sigma>,\<Gamma> \<turnstile> \<Phi> \<approx> \<Phi>' : thy; \<phi> \<in> set \<Phi>\<rbrakk> \<Longrightarrow> \<exists> \<phi>'. \<phi>' \<in> set \<Phi>' \<and>  \<Sigma>,\<Gamma> \<turnstile> \<phi> \<approx> \<phi>' : o"
  apply(induct rule: wf_equiv_thy.inducts)
  by(auto)

lemma solve_closed_under_equiv_2:
  shows "\<lbrakk>\<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup> ATOM A; \<Sigma>,\<Gamma> \<turnstile> A \<approx> A' : prop_ty; valid_ctx \<Gamma>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup> ATOM A'"
  and "\<lbrakk>\<Sigma>,\<Gamma>,\<Phi> | \<phi> \<rightharpoonup> A; \<Sigma>,\<Gamma> \<turnstile> A \<approx> A' : prop_ty; valid_ctx \<Gamma>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> | \<phi> \<rightharpoonup> A'"
  apply(nominal_induct \<Sigma> \<Gamma> \<Phi> \<phi>\<equiv>"ATOM A" and \<Sigma> \<Gamma> \<Phi> \<phi> A avoiding: A A' rule:solve_atomic_solve.strong_inducts)
    (* case 1 *)
  apply(erule solve_ATOM)
  apply(simp add: hoas_form.inject)
    (* case 2 *)
  apply(clarsimp)
    (* case 3 *)
  apply(clarsimp)
  (* case 4 *)
  apply(rule asolve_ATOM)
  apply(simp,blast intro: wf_equiv_sym wf_equiv_trans)
    (* case 5 *)
  apply(rule asolve_IMP)
  apply(blast)
  apply(blast)
    (* case 6 *)
   apply(rule asolve_FORALL, simp_all add:fresh_prod)
   apply(blast)
   done
 
lemma solve_closed_under_equiv_:
  shows "\<lbrakk>\<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup> \<phi>; \<Sigma>,\<Gamma> \<turnstile> \<phi> \<approx> \<phi>' : o; \<Sigma>,\<Gamma> \<turnstile> \<Phi> \<approx> \<Phi>' : thy; valid_ctx \<Gamma>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi>' \<rightharpoonup> \<phi>'"
  and "\<lbrakk>\<Sigma>,\<Gamma>,\<Phi> | \<phi> \<rightharpoonup> A; \<Sigma>,\<Gamma> \<turnstile> \<phi> \<approx> \<phi>' : o; \<Sigma>,\<Gamma> \<turnstile> A \<approx> A' : prop_ty; \<Sigma>,\<Gamma> \<turnstile> \<Phi> \<approx> \<Phi>' : thy; valid_ctx \<Gamma>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi>' | \<phi>' \<rightharpoonup> A'"
  apply(nominal_induct \<Sigma> \<Gamma> \<Phi> \<phi> and \<Sigma> \<Gamma> \<Phi> \<phi> A avoiding: \<phi>'  \<Phi>' A' rule:solve_atomic_solve.strong_inducts)
  (* case 1 *)
  apply(erule wf_equiv_form.cases, simp_all add:hoas_form.inject,clarify)
  apply(frule in_closed_under_equiv,simp,clarify)
  apply(erule solve_ATOM)
  apply(blast)
  (* case 2 *)
  apply(erule wf_equiv_form.cases, simp_all add:hoas_form.inject,clarify)
  apply(rule solve_IMP)
  apply(blast intro: wf_equiv_thy.intros)
    (* case 3 *)
  apply(erule_tac x="x" in  wf_equiv_form.strong_cases,simp_all add:hoas_form.inject abs_fresh alpha, clarify)
  apply(rule solve_FORALL, simp_all add:fresh_prod)
  apply(drule_tac \<Gamma>'="(x, \<tau>') # \<Gamma>'" in wf_equiv_thy_weakening)
  apply(auto)
    (* case 4 *)
  apply(erule wf_equiv_form.cases, simp_all add:hoas_form.inject,clarify)
  apply(rule asolve_ATOM)
  apply(simp,blast intro: wf_equiv_sym wf_equiv_trans)
    (* case 5 *)
  apply(erule wf_equiv_form.cases, simp_all add:hoas_form.inject,clarify)
  apply(rule asolve_IMP)
  apply(blast)
  apply(blast)
    (* case 6 *)
   apply(erule_tac x="x" in wf_equiv_form.strong_cases, simp_all add:hoas_form.inject abs_fresh alpha,clarify)
   apply(rule asolve_FORALL, simp_all add:fresh_prod)
   apply(subgoal_tac "\<Sigma>',\<Gamma>' \<turnstile> \<phi>''{x::=t} \<approx> \<phi>'a{x::=t} : o")
   prefer 2      apply(rule wf_equiv_form_subst1,auto) 
   done
 
theorem solve_closed_under_equiv:
  "\<lbrakk>\<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup> \<phi>; \<Sigma>,\<Gamma> \<turnstile> \<phi> \<approx> \<phi>' : o; \<Sigma>,\<Gamma> \<turnstile> \<Phi> : thy; valid_ctx \<Gamma>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup> \<phi>'"
  by(auto intro:  solve_closed_under_equiv_ wf_equiv_thy_refl)




text {* Height-bounded derivations *}

inductive solve_n :: "hoas_sig \<Rightarrow> hoas_ctx \<Rightarrow> hoas_thy \<Rightarrow> hoas_form \<Rightarrow> nat \<Rightarrow> bool"
  ("_,_,_ \<rightharpoonup> _ @ _" [60,60,60,60,60] 60)
and atomic_solve_n :: "hoas_sig \<Rightarrow> hoas_ctx \<Rightarrow> hoas_thy \<Rightarrow> hoas_form \<Rightarrow> hoas \<Rightarrow> nat \<Rightarrow> bool"
  ("_,_,_ | _ \<rightharpoonup> _ @ _" [60,60,60,60,60,60] 60)
where
  solven_ATOM: "\<lbrakk>\<phi> \<in> set \<Phi>; \<Sigma>,\<Gamma>,\<Phi> | \<phi> \<rightharpoonup> A @ n\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup> ATOM A @ Suc n"
| solven_IMP: "\<Sigma>,\<Gamma>,\<phi>#\<Phi> \<rightharpoonup> \<psi> @ n \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup> \<phi> \<rightarrow> \<psi> @ n"
| solven_FORALL: "\<lbrakk>\<Sigma>,(x,\<tau>)#\<Gamma>,\<Phi> \<rightharpoonup> \<phi> @ n;x \<sharp> (\<Phi>,\<Gamma>)\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup> \<Pi>[x:\<tau>].\<phi> @ n"
| asolven_ATOM: "\<lbrakk> \<Sigma>,\<Gamma> \<turnstile> A \<approx> B : prop_ty\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> | ATOM A \<rightharpoonup> B @ n"
| asolven_IMP: "\<lbrakk>\<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup> \<phi> @ n; \<Sigma>,\<Gamma>,\<Phi> | \<psi> \<rightharpoonup> A @ n\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> | \<phi> \<rightarrow> \<psi> \<rightharpoonup> A @ n"
| asolven_FORALL: "\<lbrakk>\<Sigma>,\<Gamma> \<turnstile> t : \<tau>; \<Sigma>,\<Gamma>,\<Phi> | \<phi>{x::=t} \<rightharpoonup> A@ n; x \<sharp> (\<Gamma>,\<Phi>,A)\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> | \<Pi>[x:\<tau>].\<phi> \<rightharpoonup> A @ n"

equivariance solve_n

nominal_inductive solve_n
  by(simp_all add:fresh_sig abs_fresh fresh_nat)

inductive_cases solven_inv:
"\<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup> ATOM A @ Suc n"
"\<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup> \<phi> \<rightarrow> \<psi> @ n"
"\<Sigma>,\<Gamma>,\<Phi> | ATOM A \<rightharpoonup> B @ n"
"\<Sigma>,\<Gamma>,\<Phi> | \<phi> \<rightarrow> \<psi> \<rightharpoonup> A @ n"
"\<Sigma>,\<Gamma>,\<Phi> | \<Pi>[x:\<tau>].\<phi> \<rightharpoonup> A @ n"


lemma solve_n_solve:
  "\<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup> \<phi> @ n \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup>  \<phi> "
  "\<Sigma>,\<Gamma>,\<Phi> | \<phi> \<rightharpoonup> A @ n \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> | \<phi> \<rightharpoonup> A "
  apply(nominal_induct \<Sigma> \<Gamma> \<Phi> \<phi> n and \<Sigma> \<Gamma> \<Phi> \<phi> A n rule:solve_n_atomic_solve_n.strong_inducts)
  apply(auto intro!:solve_atomic_solve.intros simp add:fresh_prod)
  done

lemma trivial:
  "\<lbrakk>Suc n \<le> m\<rbrakk> \<Longrightarrow> \<exists>m0. m = Suc m0"
  apply(arith)
  done

lemma solve_n_monotone:
   "\<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup> \<phi> @ n \<Longrightarrow> m \<ge> n \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup>  \<phi> @ m"
  "\<Sigma>,\<Gamma>,\<Phi> | \<phi> \<rightharpoonup> A @ n \<Longrightarrow>  m \<ge> n \<Longrightarrow>\<Sigma>,\<Gamma>,\<Phi> | \<phi> \<rightharpoonup> A @ m"
  apply(nominal_induct \<Sigma> \<Gamma> \<Phi> \<phi> n and \<Sigma> \<Gamma> \<Phi> \<phi> A n arbitrary: m and m rule:solve_n_atomic_solve_n.strong_inducts)
  apply(auto intro!:solve_n_atomic_solve_n.intros simp add:fresh_prod)
  apply(frule trivial, clarify)
  apply(auto intro!:solve_n_atomic_solve_n.intros simp add:fresh_prod)
  done
 

lemma solve_solve_n:
  "\<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup> \<phi> \<Longrightarrow> \<exists> n. \<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup>  \<phi> @ n"
  "\<Sigma>,\<Gamma>,\<Phi> | \<phi> \<rightharpoonup> A  \<Longrightarrow> \<exists> n. \<Sigma>,\<Gamma>,\<Phi> | \<phi> \<rightharpoonup> A @ n"
  apply(nominal_induct \<Sigma> \<Gamma> \<Phi> \<phi> and \<Sigma> \<Gamma> \<Phi> \<phi> A  rule:solve_atomic_solve.strong_inducts)
  apply(auto intro!:solve_n_atomic_solve_n.intros simp add:fresh_prod)
  apply(rule_tac x="max x xa" in exI)
  apply(rule asolven_IMP)
  apply(erule solve_n_monotone,arith)+
  done

theorem solve_equals_solve_n:
  "\<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup> \<phi> = (\<exists> n. \<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup>  \<phi> @ n)"
  "\<Sigma>,\<Gamma>,\<Phi> | \<phi> \<rightharpoonup> A  = (\<exists> n. \<Sigma>,\<Gamma>,\<Phi> | \<phi> \<rightharpoonup> A @ n)"
  apply(auto intro:solve_solve_n solve_n_solve)
  done


text {* Uniform calculus where we must apply right rules if goal is not atomic *}

inductive unif :: "hoas_sig \<Rightarrow> hoas_ctx \<Rightarrow> hoas_thy \<Rightarrow> hoas_form \<Rightarrow> bool"
  ("_,_,_ ⇛ _ unif" [60,60,60,60] 60)
where
  unif_ATOM: "\<lbrakk>ATOM B \<in> set \<Phi>; \<Sigma>,\<Gamma> \<turnstile> B \<approx> A : prop_ty\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> ⇛ ATOM A unif"
| unif_IMP_r: "\<Sigma>,\<Gamma>,\<phi>#\<Phi> ⇛ \<psi> unif \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> ⇛ \<phi> \<rightarrow> \<psi> unif"
| unif_FORALL_r: "\<lbrakk>\<Sigma>,(x,\<tau>)#\<Gamma>,\<Phi> ⇛ \<phi> unif;x \<sharp> (\<Phi>,\<Gamma>)\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> ⇛ \<Pi>[x:\<tau>].\<phi> unif"
| unif_IMP_l: "\<lbrakk>\<psi>1 \<rightarrow> \<psi>2 \<in> set \<Phi>; \<Sigma>,\<Gamma>,\<Phi> ⇛ \<psi>1 unif;  \<Sigma>,\<Gamma>,\<psi>2#\<Phi>  ⇛ ATOM A unif\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> ⇛ ATOM A unif"
| unif_FORALL_l: "\<lbrakk>\<Pi>[x:\<tau>].\<psi> \<in> set \<Phi>; \<Sigma>,\<Gamma> \<turnstile> t : \<tau>; \<Sigma>,\<Gamma>,(\<psi>{x::=t})#\<Phi> ⇛ ATOM A unif; x \<sharp> (\<Gamma>,\<Phi>,A)\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> ⇛ ATOM A unif"

equivariance unif

nominal_inductive unif
  by(simp_all add:fresh_sig abs_fresh fresh_nat)

lemma unif_in_redundant_:
  "\<lbrakk>\<Sigma>,\<Gamma>,\<Psi>@[\<psi>]@\<Phi> ⇛ \<phi> unif;  \<psi> \<in> set \<Phi>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Psi>@\<Phi> ⇛ \<phi> unif"
  apply(nominal_induct \<Sigma> \<Gamma> \<Phi>\<equiv>"\<Psi>@[\<psi>]@\<Phi>" \<phi> avoiding: \<Psi> \<psi> \<Phi> rule:unif.strong_inducts)
  apply(clarify)
  apply(rule unif_ATOM,simp_all,clarsimp)
  (* case 2 *)
  apply(clarify, rule unif_IMP_r)
  apply(drule_tac x="\<phi> # \<Psi>" in meta_spec)
  apply(drule_tac x="\<psi>'" in meta_spec)
  apply(drule_tac x="\<Phi>'" in meta_spec)
  apply(clarsimp)
(* case 3 *)
  apply(clarify, rule unif_FORALL_r)
  apply(blast)
  apply(simp add:fresh_prod fresh_list_append)
(* case 4 *)
  apply(clarsimp)
  apply(subgoal_tac "\<psi>1 \<rightarrow> \<psi>2 \<in> set (\<Psi> @ \<Phi>')")
  prefer 2 apply(clarsimp)
  apply(erule unif_IMP_l)
  apply(blast)
  apply(drule_tac x="\<psi>2 # \<Psi>" in meta_spec)+
  apply(clarsimp)
(* case 5 *)
  apply(clarsimp)
  apply(subgoal_tac "\<Pi>[x:\<tau>].\<psi> \<in> set (\<Psi> @ \<Phi>')")
  prefer 2 apply(clarsimp)
  apply(erule unif_FORALL_l)
  apply(blast)
  apply(drule_tac x="\<psi>{x::=t} # \<Psi>" in meta_spec)+
  apply(clarsimp, simp add:fresh_prod fresh_list_append)
  done

    

lemma unif_in_redundant:
  "\<lbrakk>\<Sigma>,\<Gamma>,\<psi>#\<Phi> ⇛ \<phi> unif;  \<psi> \<in> set \<Phi>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> ⇛ \<phi> unif"
  apply(auto intro: unif_in_redundant_[where \<Psi>="[]", simplified])
  done

lemma unif_weakening:
  "\<lbrakk>\<Sigma>,\<Gamma>,\<Psi> ⇛ \<phi> unif;  set \<Psi> \<subseteq> set \<Phi>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> ⇛ \<phi> unif"
  apply(nominal_induct \<Sigma> \<Gamma> \<Psi> \<phi> avoiding: \<Phi> rule:unif.strong_inducts)
  apply(auto intro:unif.intros)
  apply(rule unif_IMP_r)
  apply(drule_tac x="\<phi>#\<Phi>'" in meta_spec,auto)
  apply(subgoal_tac "\<psi>1 \<rightarrow> \<psi>2 \<in> set \<Phi>'")
  prefer 2 apply(auto)
  apply(erule unif_IMP_l,auto)
  apply(drule_tac x="\<psi>2 # \<Phi>'" in meta_spec,auto)
  apply(drule_tac x="\<psi>2 # \<Phi>'" in meta_spec,auto)
  apply(subgoal_tac "\<Pi>[x:\<tau>].\<psi> \<in> set \<Phi>'")
  prefer 2 apply(auto)
  apply(erule unif_FORALL_l,auto)
  apply(drule_tac x="\<psi>{x::=t} # \<Phi>'" in meta_spec,auto)
  done

lemma unif_ctx_weakening:
  "\<lbrakk>\<Sigma>,\<Gamma>,\<Phi> ⇛ \<phi> unif; x \<sharp> \<Gamma>; valid_ctx \<Gamma>\<rbrakk> \<Longrightarrow> \<Sigma>,(x,\<tau>)#\<Gamma>,\<Phi> ⇛ \<phi> unif"
  sorry


lemma unif_exchange:
  "\<Sigma>,\<Gamma>,\<psi>1 # \<psi>2 # \<Phi> ⇛ \<phi> unif \<Longrightarrow> \<Sigma>,\<Gamma>,\<psi>2 # \<psi>1 # \<Phi> ⇛ \<phi> unif"
  sorry




lemma solve_implies_unif:
  "\<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup> \<phi> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> ⇛ \<phi> unif"
  "\<Sigma>,\<Gamma>,\<Phi> | \<phi> \<rightharpoonup> A  \<Longrightarrow> \<Sigma>,\<Gamma>,\<phi>#\<Phi> ⇛ ATOM A unif"
  apply(nominal_induct \<Sigma> \<Gamma> \<Phi> \<phi> and \<Sigma> \<Gamma> \<Phi> \<phi> A  rule:solve_atomic_solve.strong_inducts)
  (* case 1 *)
  apply(erule unif_in_redundant,clarify)
(* case 2 *)
  apply(erule unif.intros)
(* case 3 *)
  apply(erule unif.intros, simp add:fresh_prod)
    (* case 4 *)
  apply(rule unif.intros,auto )
    (* case 5 *)
  apply(rule unif_IMP_l,auto)
  apply(erule unif_weakening,auto)
  apply(erule unif_weakening,auto)
    (* case 6 *)
  apply(rule unif_FORALL_l,auto)
  apply(erule unif_weakening,auto)
  apply(auto simp add:fresh_prod fresh_list_cons abs_fresh)
  done
  
lemma unif_implies_solve:
  "\<Sigma>,\<Gamma>,\<Phi> ⇛ \<phi> unif \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup> \<phi>"
  sorry

  

text {* Sequent calculus with unrestricted right-rules *}

inductive seq :: "hoas_sig \<Rightarrow> hoas_ctx \<Rightarrow> hoas_thy \<Rightarrow> hoas_form \<Rightarrow> bool"
  ("_,_,_ ⇛ _" [60,60,60,60] 60)
where
  seq_ATOM: "\<lbrakk>ATOM B \<in> set \<Phi>; \<Sigma>,\<Gamma> \<turnstile> B \<approx> A : prop_ty\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> ⇛ ATOM A"
| seq_IMP_r: "\<Sigma>,\<Gamma>,\<phi>#\<Phi> ⇛ \<psi> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> ⇛ \<phi> \<rightarrow> \<psi>"
| seq_FORALL_r: "\<lbrakk>\<Sigma>,(x,\<tau>)#\<Gamma>,\<Phi> ⇛ \<phi>;x \<sharp> (\<Phi>,\<Gamma>)\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> ⇛ \<Pi>[x:\<tau>].\<phi>"
| seq_IMP_l: "\<lbrakk>\<psi>1 \<rightarrow> \<psi>2 \<in> set \<Phi>; \<Sigma>,\<Gamma>,\<Phi> ⇛ \<psi>1;  \<Sigma>,\<Gamma>,\<psi>2#\<Phi>  ⇛ \<phi>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> ⇛ \<phi>"
| seq_FORALL_l: "\<lbrakk>\<Pi>[x:\<tau>].\<psi> \<in> set \<Phi>; \<Sigma>,\<Gamma> \<turnstile> t : \<tau>; \<Sigma>,\<Gamma>,(\<psi>{x::=t})#\<Phi> ⇛ \<phi>; x \<sharp> (\<Gamma>,\<Phi>,\<phi>)\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> ⇛ \<phi>"

equivariance seq

nominal_inductive seq
  by(simp_all add:fresh_sig abs_fresh fresh_nat)

lemma seq_in_redundant_:
  "\<lbrakk>\<Sigma>,\<Gamma>,\<Psi>@[\<psi>]@\<Phi> ⇛ \<phi>;  \<psi> \<in> set \<Phi>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Psi>@\<Phi> ⇛ \<phi>"
  apply(nominal_induct \<Sigma> \<Gamma> \<Phi>\<equiv>"\<Psi>@[\<psi>]@\<Phi>" \<phi> avoiding: \<Psi> \<psi> \<Phi> rule:seq.strong_inducts)
  apply(clarify)
  apply(rule seq_ATOM,simp_all,clarsimp)
  (* case 2 *)
  apply(clarify, rule seq_IMP_r)
  apply(drule_tac x="\<phi> # \<Psi>" in meta_spec)
  apply(drule_tac x="\<psi>'" in meta_spec)
  apply(drule_tac x="\<Phi>'" in meta_spec)
  apply(clarsimp)
(* case 3 *)
  apply(clarify, rule seq_FORALL_r)
  apply(blast)
  apply(simp add:fresh_prod fresh_list_append)
(* case 4 *)
  apply(clarsimp)
  apply(subgoal_tac "\<psi>1 \<rightarrow> \<psi>2 \<in> set (\<Psi> @ \<Phi>')")
  prefer 2 apply(clarsimp)
  apply(erule seq_IMP_l)
  apply(blast)
  apply(drule_tac x="\<psi>2 # \<Psi>" in meta_spec)+
  apply(clarsimp)
(* case 5 *)
  apply(clarsimp)
  apply(subgoal_tac "\<Pi>[x:\<tau>].\<psi> \<in> set (\<Psi> @ \<Phi>')")
  prefer 2 apply(clarsimp)
  apply(erule seq_FORALL_l)
  apply(blast)
  apply(drule_tac x="\<psi>{x::=t} # \<Psi>" in meta_spec)+
  apply(clarsimp, simp add:fresh_prod fresh_list_append)
  done

  
  
  

lemma seq_in_redundant:
  "\<lbrakk>\<Sigma>,\<Gamma>,\<psi>#\<Phi> ⇛ \<phi>;  \<psi> \<in> set \<Phi>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> ⇛ \<phi>"
  apply(auto intro: seq_in_redundant_[where \<Psi>="[]", simplified])
  done

lemma seq_weakening:
  "\<lbrakk>\<Sigma>,\<Gamma>,\<Psi> ⇛ \<phi>;  set \<Psi> \<subseteq> set \<Phi>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> ⇛ \<phi>"
  apply(nominal_induct \<Sigma> \<Gamma> \<Psi> \<phi> avoiding: \<Phi> rule:seq.strong_inducts)
  apply(auto intro:seq.intros)
  apply(rule seq_IMP_r)
  apply(drule_tac x="\<phi>#\<Phi>'" in meta_spec,auto)
  apply(subgoal_tac "\<psi>1 \<rightarrow> \<psi>2 \<in> set \<Phi>'")
  prefer 2 apply(auto)
  apply(erule seq_IMP_l,auto)
  apply(drule_tac x="\<psi>2 # \<Phi>'" in meta_spec,auto)
  apply(drule_tac x="\<psi>2 # \<Phi>'" in meta_spec,auto)
  apply(subgoal_tac "\<Pi>[x:\<tau>].\<psi> \<in> set \<Phi>'")
  prefer 2 apply(auto)
  apply(erule seq_FORALL_l,auto)
  apply(drule_tac x="\<psi>{x::=t} # \<Phi>'" in meta_spec,auto)
  done

lemma unif_implies_seq:
  "\<Sigma>,\<Gamma>,\<Phi> ⇛ \<phi> unif \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> ⇛ \<phi>"
  apply(nominal_induct \<Sigma> \<Gamma> \<Phi> \<phi>  rule:unif.strong_inducts)
  apply(auto intro:seq.intros)
  done


lemma solve_implies_seq: 
  "\<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup> \<phi> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> ⇛ \<phi>"
  apply(rule unif_implies_seq, erule solve_implies_unif)
  done

lemma unif_admits_IMP_l:
  "\<lbrakk> \<psi>1 \<rightarrow> \<psi>2 \<in> set \<Phi>;\<Sigma>,\<Gamma>,\<Phi> ⇛ \<psi>1 unif;\<Sigma>,\<Gamma>,\<psi>2 # \<Phi> ⇛ \<phi> unif; valid_ctx \<Gamma>\<rbrakk>
       \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> ⇛ \<phi> unif"
  apply(nominal_induct \<phi> avoiding: \<psi>1 \<psi>2 \<Phi> \<Sigma> \<Gamma> 
    rule:hoas_form.strong_inducts)
  apply(blast intro: unif_IMP_l)
  apply(erule unif.cases,simp_all add:hoas_form.inject,clarify) back
  apply(rule unif_IMP_r)
  apply(drule unif_exchange)
  apply(subgoal_tac "\<Sigma>',\<Gamma>',\<phi>#\<Phi> ⇛ \<psi>1 unif")
  prefer 2 apply(erule unif_weakening,auto)
  apply(subgoal_tac "\<psi>1 \<rightarrow> \<psi>2 \<in> set (\<phi> #\<Phi>)",blast,auto)
    (* case 3 *)
  apply(erule_tac x="var" in unif.strong_cases,simp_all add:hoas_form.inject abs_fresh alpha fresh_prod fresh_list_cons) back
  apply(auto)
  apply(rule unif_FORALL_r, simp_all add:fresh_prod fresh_list_cons)
  apply(subgoal_tac "\<Sigma>',(var, \<tau>)#\<Gamma>',\<Phi> ⇛ \<psi>1 unif")
  prefer 2 apply(erule unif_ctx_weakening,simp_all)
  apply(blast)
  done
  

lemma subst_form_fresh:
  fixes x::var
  and \<phi>::hoas_form
  shows "\<lbrakk>x \<sharp> \<phi>; x \<sharp> t\<rbrakk> \<Longrightarrow> x \<sharp> \<phi>{y::=t}"
  sorry

lemma unif_admits_FORALL_l:
  "\<lbrakk>\<Pi>[x:\<tau>].\<psi> \<in> set \<Phi>; \<Sigma>,\<Gamma> \<turnstile> t : \<tau>; \<Sigma>,\<Gamma>,\<psi>{x::=t} # \<Phi> ⇛ \<phi> unif; x \<sharp> (\<Gamma>,\<Phi>,\<phi>); valid_ctx \<Gamma>\<rbrakk>
       \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> ⇛ \<phi> unif"
  apply(nominal_induct \<phi> avoiding: \<psi> \<Phi> \<Sigma> \<Gamma> x t
    rule:hoas_form.strong_inducts)
  apply(erule unif_FORALL_l)
  apply(assumption)
  apply(assumption)
  apply(simp add:fresh_prod)
  (* case 2 *)
  apply(erule unif.cases,simp_all add:hoas_form.inject,clarify)
  apply(rule unif_IMP_r)
  apply(simp add:fresh_prod,clarify)
  apply(drule unif_exchange)
  apply(subgoal_tac "\<Pi>[x:\<tau>].\<psi> \<in> set (\<phi> # \<Phi>)")
  apply(subgoal_tac "x \<sharp> (\<phi> # \<Phi>)")
  apply(blast)
  apply(simp add:fresh_prod)
  apply(simp add:fresh_list_cons)
  apply(clarsimp)
  (* case 3 *)
  apply(erule_tac x="var" in unif.strong_cases,simp_all add:hoas_form.inject abs_fresh alpha fresh_prod fresh_list_cons subst_form_fresh fresh_atm) 
  apply(clarify)
  apply(simp add:fresh_list_cons)
  apply(rule unif_FORALL_r)
  apply(drule_tac x="\<psi>" in meta_spec )
  apply(drule_tac x="\<Phi>" in meta_spec )
  apply(drule_tac x="\<Sigma>'" in meta_spec)
  apply(drule_tac x="(var,\<tau>')#\<Gamma>'" in meta_spec)
  apply(drule_tac x=x in meta_spec)
  apply(drule_tac x="t" in meta_spec)
  apply(simp_all add:fresh_prod fresh_list_cons fresh_atm)
  apply(drule meta_mp)
  apply(erule weakening,auto)
  
  apply(auto intro: subst_form_fresh)
  done

lemma seq_implies_unif:
  "\<Sigma>,\<Gamma>,\<Phi> ⇛ \<phi> \<Longrightarrow> valid_ctx \<Gamma> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> ⇛ \<phi> unif"
  apply(nominal_induct \<Sigma> \<Gamma> \<Phi> \<phi> rule:seq.strong_inducts)
  apply(auto intro:unif.intros unif_admits_IMP_l unif_admits_FORALL_l)
  apply(subgoal_tac "valid_ctx ((x, \<tau>) # \<Gamma>)",auto intro:unif.intros unif_admits_IMP_l unif_admits_FORALL_l)
  done  

 

lemma seq_implies_solve:
  "\<Sigma>,\<Gamma>,\<Phi> ⇛ \<phi> \<Longrightarrow> valid_ctx \<Gamma> \<Longrightarrow> \<Sigma>,\<Gamma>,\<Phi> \<rightharpoonup> \<phi>"
  apply(rule unif_implies_solve,erule seq_implies_unif, simp)
  done



    
end

