theory Signature
imports HOAS Logic
begin

text {* Defines signature of HOAS for lambda terms *}

constdefs cLAM :: string
"cLAM == ''lam''"
constdefs cAPP :: string
"cAPP == ''app''"
constdefs cRED :: string
"cRED == ''red''"
constdefs cEXP :: string
"cEXP == ''exp''"

lemma [simp]: 
  "supp(cLAM) = {}"
  "supp(cAPP) = {}"
  "supp(cRED) = {}"
  "supp(cEXP) = {}"
  by(simp_all add:supp_string)

lemma [simp]: 
  fixes x::var
  shows "x \<sharp> cLAM"
  and  "x \<sharp> cAPP"
  and  "x \<sharp> cEXP"
  and  "x \<sharp> cRED"
  by(simp_all add:fresh_def)

lemma [eqvt]: 
  "p \<bullet> cLAM = cLAM"
  "p \<bullet> cAPP = cAPP"
  "p \<bullet> cRED = cRED"
  "p \<bullet> cEXP = cEXP"
  by(simp_all  add:perm_string)


constdefs exp_ty :: "hoas_ty"
"exp_ty \<equiv> BASE cEXP"
constdefs app_exp :: "hoas \<Rightarrow> hoas \<Rightarrow> hoas"
"app_exp m n == APP (APP (CST cAPP) m) n"
constdefs lam_exp :: "hoas => hoas" 
"lam_exp f == APP (CST cLAM) f"
constdefs red_prop :: "hoas \<Rightarrow> hoas \<Rightarrow> hoas_form"
"red_prop m n \<equiv> ATOM (APP (APP (CST cRED) m) n)"


lemma [eqvt]: 
  fixes p::"var prm"
  shows "p \<bullet> exp_ty = exp_ty"
  and "p \<bullet> (app_exp m n) = app_exp (p \<bullet> m) (p \<bullet> n)"
  and "p \<bullet> (lam_exp m) = lam_exp (p \<bullet> m)"
  and "p \<bullet> (red_prop m n) = red_prop (p \<bullet> m) (p \<bullet> n)"
  by (simp_all add: exp_ty_def app_exp_def lam_exp_def red_prop_def perm_string eqvts)

lemma [intro]: 
  fixes x::"var"
  shows "x \<sharp> exp_ty"
  and "x \<sharp> m \<and> x \<sharp> n \<Longrightarrow> x \<sharp> (app_exp m n)"
  and "x \<sharp> m \<Longrightarrow> x \<sharp> (lam_exp m)"
  apply (auto simp add:app_exp_def lam_exp_def)
  done


lemma  subst_exp[simp]:
  "(app_exp m n){x::=u} = app_exp (m{x::=u}) (n{x::=u})"
  "(lam_exp m){x::=u} = lam_exp (m{x::=u})"
  apply(auto simp add:app_exp_def lam_exp_def)
  done


lemmas sig_lam_simps[simp] = exp_ty_def app_exp_def lam_exp_def red_prop_def 




constdefs sig_lam :: "hoas_sig"
"sig_lam == [(cLAM,(exp_ty ~> exp_ty) ~> exp_ty),
             (cAPP,exp_ty ~> (exp_ty ~> exp_ty)),
             (cRED, exp_ty ~> (exp_ty ~> prop_ty))]"

lemma sig_lam_lookup : 
  shows "(cAPP,ty) \<in> set sig_lam \<Longrightarrow> ty = exp_ty ~> (exp_ty ~> exp_ty)"
  and "(cLAM,ty) \<in> set sig_lam \<Longrightarrow> ty = (exp_ty ~> exp_ty) ~> exp_ty"
  and "(cRED,ty) \<in> set sig_lam \<Longrightarrow> ty = exp_ty ~> (exp_ty ~> prop_ty)"
  by(auto simp add:sig_lam_def cAPP_def cLAM_def cRED_def)

lemma sig_lam_elem_: 
  shows "(cAPP,exp_ty ~> (exp_ty ~> exp_ty)) \<in> set sig_lam"
  and "(cRED,exp_ty ~> (exp_ty ~> prop_ty)) \<in> set sig_lam"
  and "(cLAM,(exp_ty ~> exp_ty) ~> exp_ty) \<in> set sig_lam"
  by(auto simp add:sig_lam_def cAPP_def cLAM_def cRED_def)

lemmas sig_lam_elem = sig_lam_elem_[simplified]
  
lemma sig_lam_valid: "valid_sig sig_lam"
  apply(auto simp add: sig_lam_def )
  apply(intro valid_sig_cons valid_sig_nil)
  apply(simp_all add:cLAM_def cAPP_def cRED_def)
  done

(* Theory lambda-calculus reduction *)

constdefs RED_app1:: "hoas_form"
  "RED_app1 \<equiv> \<Pi>[var 0:exp_ty]. (
              \<Pi>[var 1:exp_ty]. (
              \<Pi>[var 2:exp_ty]. (
  red_prop (VAR (var 0)) (VAR (var 2)) \<rightarrow> red_prop (app_exp (VAR (var 0)) (VAR (var 1))) (app_exp (VAR (var 2)) (VAR (var 1))))))"

constdefs RED_app2:: "hoas_form"
  "RED_app2 \<equiv> \<Pi>[var 0:exp_ty]. (
              \<Pi>[var 1:exp_ty]. (
              \<Pi>[var 2:exp_ty]. (
  red_prop (VAR (var 1)) (VAR (var 2)) \<rightarrow> red_prop (app_exp (VAR (var 0)) (VAR (var 1))) (app_exp (VAR (var 0)) (VAR (var 2))))))"

constdefs RED_lam:: "hoas_form"
"RED_lam \<equiv> 
\<Pi>[var 0:(exp_ty ~> exp_ty)].( 
\<Pi>[var 1:(exp_ty ~> exp_ty)].( 
  (\<Pi>[var 2:exp_ty]. (red_prop (APP (VAR (var 0)) (VAR (var 2))) (APP (VAR (var 1)) (VAR (var 2))))) \<rightarrow> red_prop (lam_exp (VAR (var 0))) (lam_exp (VAR (var 1)))))"


constdefs RED_beta::"hoas_form"
"RED_beta \<equiv> \<Pi>[var 0: (exp_ty ~> exp_ty)]. ( 
            \<Pi>[var 1: exp_ty]. (
            red_prop (app_exp (lam_exp (VAR (var 0))) (VAR (var 1))) (APP (VAR (var 0)) (VAR (var 1)))))"


constdefs thy_lam :: "hoas_thy"
"thy_lam \<equiv> [
  RED_app1,
  RED_app2,
  RED_lam,
  RED_beta
]"


lemma RED_wf:
  "sig_lam, [] \<turnstile> RED_app1 : o"
  "sig_lam, [] \<turnstile> RED_app2 : o"
  "sig_lam, [] \<turnstile> RED_lam : o"
  "sig_lam, [] \<turnstile> RED_beta : o"
  apply(auto intro!:wf_form.intros sig_lam_elem 
    simp add:RED_app1_def RED_app2_def RED_lam_def RED_beta_def 
    red_prop_def lam_exp_def app_exp_def 
    fresh_list_nil fresh_list_cons fresh_ty fresh_prod fresh_atm)
  done

lemma thy_lam_wf: "valid_ctx \<Gamma> \<Longrightarrow> sig_lam, \<Gamma> \<turnstile> thy_lam : thy"
  apply(subgoal_tac "sig_lam, [] \<turnstile> thy_lam : thy")
  apply(erule_tac \<Gamma>="[]" in  wf_thy_weakening)
  apply(auto simp add:thy_lam_def intro!: wf_thy.intros RED_wf)
  done
  

lemma thy_lam_disj: 
  "\<phi> \<in> set thy_lam \<Longrightarrow> \<phi> = RED_app1 \<or> \<phi> = RED_app2 \<or> \<phi> = RED_lam \<or> \<phi> = RED_beta"
  apply(simp add:thy_lam_def)
done

lemma thy_lam_fresh:
  fixes x::var 
  shows "x \<sharp> thy_lam"
  apply(simp add:thy_lam_def fresh_list_cons fresh_list_nil abs_fresh RED_app1_def RED_app2_def RED_lam_def RED_beta_def) 
  apply(case_tac "x=var 0",simp_all add:fresh_atm)
  apply(case_tac "x=var 1",simp_all add:fresh_atm)
  done


end
  