theory Translation
imports Nominal Name Lambda HOAS Logic  Signature

begin


constdefs trans_ctx :: "var list \<Rightarrow> hoas_ctx"
"trans_ctx X == map (\<lambda> x. (x,exp_ty)) X"

lemma ctx_lookup : "(v, ty) \<in> set (trans_ctx X) \<Longrightarrow> ty = exp_ty"
  apply(induct X)
  apply(auto simp add:trans_ctx_def)
  done

lemma trans_ctx_fresh: 
  fixes x::"var"
  shows "x \<sharp> X \<Longrightarrow> x \<sharp> trans_ctx X"  
  apply(induct X arbitrary: x)
  apply(auto simp add:trans_ctx_def fresh_list_nil fresh_list_cons fresh_prod fresh_atm)
  done


lemma trans_ctx_fresh_inv: 
  fixes x::"var"
  shows "x \<sharp> trans_ctx X \<Longrightarrow> x \<sharp> X"  
  apply(induct X arbitrary: x)
  apply(auto simp add:trans_ctx_def fresh_list_nil fresh_list_cons fresh_prod fresh_atm)
  done

lemma trans_ctx_simp:
  "trans_ctx [] = []"
  "trans_ctx (x#X) = (x,BASE cEXP)#(trans_ctx X)"
  apply(auto simp add:trans_ctx_def)
  done

lemma valid_trans_valid: "valid X \<Longrightarrow> valid_ctx (trans_ctx X)"
  apply(induct X rule:valid.inducts)
  apply(auto simp add:trans_ctx_simp trans_ctx_fresh)
  done


lemma trans_set: "x \<in> set X \<Longrightarrow> (x, exp_ty) \<in> set (trans_ctx X)"
  by(induct X)
  (auto simp add:trans_ctx_simp)

lemma trans_set_inv: " (x, exp_ty) \<in> set (trans_ctx X) \<Longrightarrow> x \<in> set X"
  by(induct X)
  (auto simp add:trans_ctx_simp)

text {* The encoding function *}

nominal_primrec 
enc :: "lam \<Rightarrow> hoas"
where
  [simplified]: "enc (Var x) = (VAR x)"
| [simplified]:"enc (App t1 t2) = app_exp (enc t1) (enc t2)"
| [simplified]:"enc (Lam x t) = lam_exp (LAM x (enc t))"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(simp add:abs_fresh)
  apply(simp add:fresh_string)
  apply(fresh_guess)+
  done

lemma enc_eqvt[eqvt]: 
  fixes p::"var prm"
  shows "p \<bullet> (enc t) = enc (p \<bullet> t)"
  apply(nominal_induct t rule:lam.strong_induct)
  apply(simp_all add:perm_string eqvts)
  done

lemma enc_fresh : "(x::var) \<sharp> t \<Longrightarrow> x \<sharp> enc t"
  apply(nominal_induct t avoiding: x rule:lam.strong_induct)
  apply(auto simp add: abs_fresh)
  done

lemma enc_fresh_inv: 
  fixes x::var
  shows "x \<sharp> enc t \<Longrightarrow> x \<sharp> t"
  apply(nominal_induct t avoiding: x rule:lam.strong_induct)
  apply(auto simp add: abs_fresh)
  done


text {* Characterization of range of enc *}

inductive enc_range :: "hoas \<Rightarrow> bool"
where
  enc_var[intro!] : "enc_range (VAR x)"
| enc_app[intro!] : "\<lbrakk>enc_range u1 ; enc_range u2 \<rbrakk> \<Longrightarrow> enc_range (APP (APP (CST cAPP) u1) u2)"
| enc_lam[intro!] : "\<lbrakk>enc_range u\<rbrakk> \<Longrightarrow> enc_range (APP (CST cLAM) (LAM x u))"

equivariance enc_range

nominal_inductive enc_range
  apply(auto simp add:hoas.fresh fresh_string abs_fresh)
  done


inductive_cases enc_range_inv[elim!]:
  "enc_range (VAR x)"
  "enc_range (app_exp u1 u2)"
  "enc_range (lam_exp (LAM x u))"





lemma enc_in_range : "enc_range(enc t)"
  apply(nominal_induct t rule:lam.strong_induct)
  apply(auto simp add:app_exp_def lam_exp_def)+
  done
 

lemma range_in_enc : "enc_range u \<Longrightarrow> \<exists> t. enc t = u"
  apply(erule_tac enc_range.induct)
  apply(rule_tac x="Var x" in exI)
  apply(simp)
  apply(erule_tac exE)+
  apply(rule_tac x="App t ta" in exI)
  apply(simp add:app_exp_def)
  apply(erule_tac exE)+
  apply(rule_tac x="Lam x t" in exI)
  apply(simp add:lam_exp_def)
  done

text {* enc is surjective on enc_range *}

theorem enc_surjective : "enc ` UNIV = {u. enc_range u}"
  apply(rule equalityI)
  apply(clarsimp,simp add:enc_in_range)
  apply(clarsimp, simp add:image_def)
  apply(frule range_in_enc)
  apply(blast)
  done




text {* Definition of the translation, relationally *}

inductive trans :: "lam \<Rightarrow> hoas \<Rightarrow> bool" (" _ ~ _" [60,60] 60)
where
  trans_var[intro!]: "(Var x) ~ (VAR x)"
| trans_app[intro!]: "\<lbrakk>t1 ~ u1 ; t2 ~ u2\<rbrakk> \<Longrightarrow> 
              App t1 t2 ~ APP (APP (CST cAPP) u1) u2"
| trans_lam[intro!]: "\<lbrakk>t~u\<rbrakk> \<Longrightarrow>
              Lam x t ~ APP (CST cLAM) (LAM x u)"

equivariance trans

nominal_inductive trans
  apply(auto simp add:abs_fresh)
  done

inductive_cases trans_inv[elim!]:
  "Var x ~ u"
  "t ~ VAR y"
  "App t1 t2 ~ u"
  "t ~ APP (APP (CST cAPP) u1) u2"
  "Lam x t ~ u"
  "t ~ APP (CST cLAM) (LAM x u)"

lemma trans_rev_fresh_ : "t ~ u \<Longrightarrow> ((a::var) \<sharp> u \<longrightarrow>  a \<sharp> t)"
  apply(erule trans.induct)
  apply(auto simp add:hoas.fresh abs_fresh app_exp_def lam_exp_def)
  done

lemma trans_fresh_ : "t ~ u \<Longrightarrow> ((a::var) \<sharp> t \<longrightarrow>  a \<sharp> u)"
  apply(erule trans.induct)
  apply(auto simp add:hoas.fresh abs_fresh lam.fresh fresh_string app_exp_def lam_exp_def)
  done

lemma trans_fresh : "t ~ u \<Longrightarrow> ((a::var) \<sharp> t = a \<sharp> u)"
  apply(auto)
  apply(drule_tac a="a" in  trans_fresh_,simp)
  apply(drule_tac a="a" in trans_rev_fresh_,simp)
  done



text {* Graph of enc is contained in trans *}

lemma enc_trans : "t ~ enc t"
  apply(nominal_induct t rule:lam.strong_induct)
  apply(auto simp add:app_exp_def lam_exp_def)
  done
  
text {* injective on enc_range *}


lemma trans_range_unique : "\<lbrakk>t ~ u\<rbrakk> \<Longrightarrow> enc t = u"
  apply(nominal_induct t u rule: trans.strong_inducts)
  apply(clarsimp)
  apply(erule trans.cases,simp_all add:hoas.inject app_exp_def lam_exp_def)
  done

lemma graph_enc_is_trans: "(enc t = u) = t ~ u"
  apply(auto intro: enc_trans trans_range_unique)
  done

lemma trans_inj: "\<lbrakk>t ~ u; t' ~ u\<rbrakk> \<Longrightarrow> t = t'"
  apply(nominal_induct t u avoiding: t' rule:trans.strong_inducts)
  apply(clarify, simp add:hoas.inject)
  apply(erule trans_inv, simp_all (no_asm_use) add:hoas.inject app_exp_def lam_exp_def)
  apply(clarify,blast)
  apply(erule_tac x="x" in  trans.strong_cases,simp_all  (no_asm_use) add:hoas.inject abs_fresh alpha) 
  apply(drule meta_mp,simp)
  apply(clarify,blast)
  done

lemma enc_inj: "inj enc"
  apply(simp add:inj_on_def)
  apply(clarify)
  apply(simp add:graph_enc_is_trans)
  apply(erule trans_inj)
  apply(rule enc_trans)
  done


  


text {* Definition of sets of expressions over a translated context *}

constdefs EXP_in :: "var list \<Rightarrow> hoas set" ("EXP[_]")
"EXP[X] == {M.  sig_lam, trans_ctx X \<turnstile> M : exp_ty}"
constdefs EXP :: "hoas set"
"EXP == EXP[[]]"

constdefs EQ_in :: "var list \<Rightarrow> (hoas * hoas) set" ("EQ[_]")
"EQ[X] == {(M,N).  sig_lam, trans_ctx X \<turnstile> M \<approx> N : exp_ty}"
constdefs EQ :: "(hoas*hoas) set"
"EQ == EQ[[]]"

constdefs RED_in :: "var list \<Rightarrow> (hoas * hoas) set" ("RED[_]")
"RED[X] == {(M,N). sig_lam,(trans_ctx X),thy_lam \<rightharpoonup> red_prop M N }"
constdefs RED :: "(hoas * hoas) set"
"RED == RED[[]]"





     

end
